@@ -241,7 +241,7 @@ namespace System.Linq.Expressions {
 				}
 
 				// Use IsNumber to avoid expensive reflection.
-				if (IsNumber (ultype)){
+				if (IsNumber (ultype)) {
 					if (ultype == urtype && ltype == rtype)
 						return null;
 
@@ -265,6 +265,9 @@ namespace System.Linq.Expressions {
 					if (!ltype.IsValueType && !rtype.IsValueType)
 						return null;
 
+					if (ltype == rtype && ultype.IsEnum)
+						return null;
+
 					if (ltype == rtype && ultype == typeof (bool))
 						return null;
 				}
@@ -983,6 +986,8 @@ namespace System.Linq.Expressions {
 			if (expression == null)
 				throw new ArgumentNullException ("expression");
 
+			CheckNonGenericMethod (propertyAccessor);
+
 			var prop = GetAssociatedProperty (propertyAccessor);
 			if (prop == null)
 				throw new ArgumentException ("propertyAccessor");
@@ -1062,7 +1067,7 @@ namespace System.Linq.Expressions {
 			return new MethodCallExpression (instance, method, args);
 		}
 
-		static bool MethodMatch (MethodInfo method, string name, Type [] parameterTypes)
+		static bool MethodMatch (MethodInfo method, string name, Type [] parameterTypes, Type [] argumentTypes)
 		{
 			if (method.Name != name)
 				return false;
@@ -1072,8 +1077,13 @@ namespace System.Linq.Expressions {
 			if (parameters.Length != parameterTypes.Length)
 				return false;
 
-			if (method.IsGenericMethod) // if it's a generic method, when can't compare its parameters
-				return true;
+			if (method.IsGenericMethod && method.IsGenericMethodDefinition) {
+				var closed = TryMakeGeneric (method, argumentTypes);
+				if (closed == null)
+					return false;
+
+				return MethodMatch (closed, name, parameterTypes, argumentTypes);
+			}
 
 			for (int i = 0; i < parameters.Length; i++)
 				if (!IsAssignableToParameterType (parameterTypes [i], parameters [i]))
@@ -1085,7 +1095,7 @@ namespace System.Linq.Expressions {
 		static MethodInfo TryGetMethod (Type type, string methodName, BindingFlags flags, Type [] parameterTypes, Type [] argumentTypes)
 		{
 			var methods = from meth in type.GetMethods (flags)
-						  where MethodMatch (meth, methodName, parameterTypes)
+						  where MethodMatch (meth, methodName, parameterTypes, argumentTypes)
 						  select meth;
 
 			if (methods.Count () > 1)
@@ -1900,8 +1910,16 @@ namespace System.Linq.Expressions {
 			return arguments.ToList ();
 		}
 
+		static void CheckNonGenericMethod (MethodBase method)
+		{
+			if (method.IsGenericMethodDefinition || method.ContainsGenericParameters)
+				throw new ArgumentException ("Can not used open generic methods");
+		}
+
 		static ReadOnlyCollection<Expression> CheckMethodArguments (MethodBase method, IEnumerable<Expression> args)
 		{
+			CheckNonGenericMethod (method);
+
 			var arguments = CreateArgumentList (args);
 			var parameters = method.GetParameters ();
 
@@ -2062,6 +2080,8 @@ namespace System.Linq.Expressions {
 			if (propertyAccessor == null)
 				throw new ArgumentNullException ("propertyAccessor");
 
+			CheckNonGenericMethod (propertyAccessor);
+
 			if (!propertyAccessor.IsStatic) {
 				if (expression == null)
 					throw new ArgumentNullException ("expression");
@@ -2204,7 +2224,7 @@ namespace System.Linq.Expressions {
 			if (IsInt (t))
 				return true;
 
-			return t == typeof (float) || t == typeof (double) || t == typeof (decimal);
+			return t == typeof (float) || t == typeof (double);
 		}
 
 		static bool IsSignedNumber (Type t)
@@ -2231,7 +2251,7 @@ namespace System.Linq.Expressions {
 		//
 		internal virtual void Emit (EmitContext ec)
 		{
-			throw new NotImplementedException ();
+			throw new NotImplementedException (String.Format ("Emit method is not implemented in expression type {0}", GetType ()));
 		}
 	}
 }
