@@ -11,10 +11,12 @@
 //    Tim Coleman <tim@timcoleman.com>
 //    Daniel Moragn <monodanmorg@yahoo.com>
 //    Hubert FONGARNAND <informatique.internet@fiducial.fr>
+//    Veerapuram Varadhan  <vvaradhan@novell.com>	
 //
 // Copyright (C) Tim Coleman , 2003
 // Copyright (C) Daniel Morgan, 2005, 2008
 // Copyright (C) Hubert FONGARNAND, 2005
+// Copyright (C) Novell Inc, 2009
 //
 // Licensed under the MIT/X11 License.
 //
@@ -123,9 +125,17 @@ namespace System.Data.OracleClient
 		{
 			this.name = name;
 			this.value = value;
+			InferOracleType (value);			
 			srcColumn = string.Empty;
 			SourceVersion = DataRowVersion.Current;
-			InferOracleType (value);
+
+#if NET_2_0
+			// Find the OciType before inferring for the size
+			if (value != null && value != DBNull.Value) {
+				this.sizeSet = true;
+				this.size = InferSize ();
+			}
+#endif
 		}
 
 		public OracleParameter (string name, OracleType oracleType)
@@ -149,11 +159,18 @@ namespace System.Data.OracleClient
 			this.name = name;
 			if (size < 0)
 				throw new ArgumentException("Size must be not be negative.");
-			this.size = size;
+			
 			this.value = value;
+			this.size = size;
+			Direction = direction;
+
+			// set sizeSet to true iff value is not-null or non-zero size value
+			if (((value != null && value != DBNull.Value) || Direction == ParameterDirection.Output) && 
+			    size > 0) 			    
+				this.sizeSet = true;
+
 			SourceColumnNullMapping = sourceColumnNullMapping;
 			OracleType = oracleType;
-			Direction = direction;
 			SourceColumn = sourceColumn;
 			SourceVersion = sourceVersion;
 		}
@@ -164,14 +181,22 @@ namespace System.Data.OracleClient
 			this.name = name;
 			if (size < 0)
 				throw new ArgumentException("Size must be not be negative.");
-			this.size = size;
+			
 			this.value = value;
+			this.size = size;
+
+			Direction = direction;
+			
+			// set sizeSet to true iff value is not-null or non-zero size value
+			if (((value != null && value != DBNull.Value) || Direction == ParameterDirection.Output) && 
+			    size > 0) 			    
+				this.sizeSet = true;
+
 			this.isNullable = isNullable;
 			this.precision = precision;
 			this.scale = scale;
 
 			OracleType = oracleType;
-			Direction = direction;
 			SourceColumn = srcColumn;
 			SourceVersion = srcVersion;
 		}
@@ -209,7 +234,11 @@ namespace System.Data.OracleClient
 #endif
 		ParameterDirection Direction {
 			get { return direction; }
-			set { direction = value; }
+			set { 
+				direction = value; 
+				if (this.size > 0 && direction == ParameterDirection.Output)
+					this.sizeSet = true;
+			}
 		}
 
 #if !NET_2_0
@@ -353,6 +382,12 @@ namespace System.Data.OracleClient
 				this.value = value;
 				if (!oracleTypeSet)
 					InferOracleType (value);
+#if NET_2_0
+				if (value != null && value != DBNull.Value) {
+					this.size = InferSize ();
+					this.sizeSet = true;
+				}
+#endif
 			}
 		}
 
@@ -441,16 +476,6 @@ namespace System.Data.OracleClient
 						svalue = svalue.ToString () + '\0';
 					}
 
-					// set bind length to size of data
-					//bindSize = (size + 1) * 4;
-					if (direction == ParameterDirection.Input)
-						bindSize = Encoding.UTF8.GetMaxByteCount (svalue.Length);
-					else
-						bindSize = Encoding.UTF8.GetMaxByteCount (size + 1);
-
-					// allocate memory based on bind length
-					bytes = new byte [bindSize];				
-
 					if (direction == ParameterDirection.Input ||
 						direction == ParameterDirection.InputOutput) {
 						
@@ -459,8 +484,13 @@ namespace System.Data.OracleClient
 						rsize = 0;
 						// Get size of buffer
 						status = OciCalls.OCIUnicodeToCharSet (statement.Parent, null, svalue, out rsize);
+
+						// allocate memory based on oracle returned length
+						bytes = new byte [rsize];
+
 						// Fill buffer
 						status = OciCalls.OCIUnicodeToCharSet (statement.Parent, bytes, svalue, out rsize);
+						bindSize = bytes.Length;
 					}
 					break;
 				case OciDataType.Date:
@@ -556,10 +586,6 @@ namespace System.Data.OracleClient
 					bindType = OciDataType.String;
 					indicator = 0;
 					svalue = "\0";
-					// define size for binding
-					bindSize = 30; // a NUMBER is 22 bytes but as a string we need more
-					// allocate memory
-					bytes = new byte [bindSize];
 					// convert value from managed type to type to marshal
 					if (direction == ParameterDirection.Input || 
 						direction == ParameterDirection.InputOutput) {
@@ -578,8 +604,8 @@ namespace System.Data.OracleClient
 						// Get size of buffer
 						OciCalls.OCIUnicodeToCharSet (statement.Parent, null, svalue, out rsize);
 
-						// Fill buffer
-						bytes = new byte [bindSize];
+						// Fill buffer 
+						bytes = new byte [rsize];
 						OciCalls.OCIUnicodeToCharSet (statement.Parent, bytes, svalue, out rsize);
 					} 
 					break;
@@ -652,6 +678,7 @@ namespace System.Data.OracleClient
 						bindValue = lobLocator.Handle;
 						lobLocator.ErrorHandle = connection.ErrorHandle;
 						lobLocator.Service = statement.Service;
+						lobLocator.Environment = connection.Environment;
 						useRef = true;
 					}
 					break;
@@ -704,6 +731,7 @@ namespace System.Data.OracleClient
 						bindValue = lobLocator.Handle;
 						lobLocator.ErrorHandle = connection.ErrorHandle;
 						lobLocator.Service = connection.ServiceContext;
+						lobLocator.Environment = connection.Environment;
 						useRef = true;
 					}
 					break;
@@ -865,6 +893,10 @@ namespace System.Data.OracleClient
 
 		private void InferOracleType (object value)
 		{
+			// Should we throw an exception here?
+			if (value == null || value == DBNull.Value)
+				return;
+			
 			Type type = value.GetType ();
 			string exception = String.Format ("The parameter data type of {0} is invalid.", type.FullName);
 			switch (type.FullName) {
