@@ -3,8 +3,10 @@
 //
 // Authors:
 //	Gonzalo Paniagua Javier (gonzalo@ximian.com)
+//      Marek Habersack <mhabersack@novell.com>
 //
 // (C) 2002,2003 Ximian, Inc (http://www.ximian.com)
+// (C) 2004-2009 Novell, Inc (http://novell.com)
 //
 
 //
@@ -52,17 +54,21 @@ namespace System.Web.Compilation
 		static readonly object textParsedEvent = new object ();
 #if NET_2_0
 		static readonly object parsingCompleteEvent = new object();
-		
-		byte[] md5checksum;
+
+		MD5 checksum;
 #endif
 		AspTokenizer tokenizer;
 		int beginLine, endLine;
 		int beginColumn, endColumn;
 		int beginPosition, endPosition;
 		string filename;
-		string fileText;
 		string verbatimID;
-
+		string fileText;
+		StringReader fileReader;
+		bool _internal;
+		int _internalLineOffset;
+		AspParser outer;
+		
 		EventHandlerList events = new EventHandlerList ();
 		
 		public event ParseErrorHandler Error {
@@ -90,23 +96,38 @@ namespace System.Web.Compilation
 		public AspParser (string filename, TextReader input)
 		{
 			this.filename = filename;
-			fileText = input.ReadToEnd ();
-#if NET_2_0
-			MD5 md5 = MD5.Create ();
-			md5checksum = md5.ComputeHash (Encoding.UTF8.GetBytes (fileText));
-#endif
-			StringReader reader = new StringReader (fileText);
-			tokenizer = new AspTokenizer (reader);
+			this.fileText = input.ReadToEnd ();
+			this.fileReader = new StringReader (this.fileText);
+			this._internalLineOffset = 0;
+			tokenizer = new AspTokenizer (this.fileReader);
 		}
 
+		public AspParser (string filename, TextReader input, int startLineOffset, AspParser outer)
+			: this (filename, input)
+		{
+			this._internal = true;
+			this._internalLineOffset = startLineOffset;
+			this.outer = outer;
+		}
+		
 #if NET_2_0
 		public byte[] MD5Checksum {
-			get { return md5checksum; }
+			get {
+				if (checksum == null)
+					return new byte[0];
+				
+				return checksum.Hash;
+			}
 		}
 #endif
 		
 		public int BeginLine {
-			get { return beginLine; }
+			get {
+				if (Internal)
+					return beginLine + InternalLineOffset;
+
+				return beginLine;
+			}
 		}
 
 		public int BeginColumn {
@@ -114,28 +135,71 @@ namespace System.Web.Compilation
 		}
 
 		public int EndLine {
-			get { return endLine; }
+			get {
+				if (Internal)
+					return endLine + InternalLineOffset;
+				return endLine;
+			}
 		}
 
 		public int EndColumn {
 			get { return endColumn; }
 		}
 
+		public bool Internal {
+			get { return _internal; }
+			set { _internal = value; }
+		}
+
+		public int InternalLineOffset {
+			get { return _internalLineOffset; }
+			set { _internalLineOffset = value; }
+		}
+		
 		public string FileText {
-			get { return fileText; }
+			get {
+				string ret = null;
+				
+				if (Internal && outer != null)
+					ret = outer.FileText;
+				
+				if (ret == null && fileText != null)
+					ret = fileText;
+				
+				return ret;
+			}
 		}
 		
 		public string PlainText {
 			get {
-				if (beginPosition >= endPosition)
+				if (beginPosition >= endPosition || fileText == null)
 					return null;
 
-				return fileText.Substring (beginPosition, endPosition - beginPosition);
+				string text = FileText;
+				int start, len;
+				
+				if (Internal && outer != null) {
+					start = beginPosition + InternalLineOffset;
+					len = (endPosition + InternalLineOffset) - start;
+				} else {
+					start = beginPosition;
+					len = endPosition - beginPosition;
+				}
+				
+				if (text != null)
+					return text.Substring (start, len);
+
+				return null;
 			}
 		}
 
 		public string Filename {
-			get { return filename; }
+			get {
+				if (Internal && outer != null)
+					return outer.Filename;
+				
+				return filename;
+			}
 		}
 
 		public string VerbatimID {
@@ -144,10 +208,11 @@ namespace System.Web.Compilation
 				verbatimID = value;
 			}
 		}
-		
+
 		bool Eat (int expected_token)
 		{
-			if (tokenizer.get_token () != expected_token) {
+			int token = tokenizer.get_token ();
+			if (token != expected_token) {
 				tokenizer.put_back ();
 				return false;
 			}
@@ -173,60 +238,76 @@ namespace System.Web.Compilation
 
 		public void Parse ()
 		{
+			if (tokenizer == null) {
+				OnError ("AspParser not initialized properly.");
+				return;
+			}
+			
 			int token;
 			string id;
 			TagAttributes attributes;
 			TagType tagtype = TagType.Text;
 			StringBuilder text =  new StringBuilder ();
 
-			while ((token = tokenizer.get_token ()) != Token.EOF) {
-				BeginElement ();
+			try {
+				while ((token = tokenizer.get_token ()) != Token.EOF) {
+					BeginElement ();
 
-				if (tokenizer.Verbatim){
-					string end_verbatim = "</" + verbatimID + ">";
-					string verbatim_text = GetVerbatim (token, end_verbatim);
+					if (tokenizer.Verbatim){
+						string end_verbatim = "</" + verbatimID + ">";
+						string verbatim_text = GetVerbatim (token, end_verbatim);
 
-					if (verbatim_text == null)
-						OnError ("Unexpected EOF processing " + verbatimID);
+						if (verbatim_text == null)
+							OnError ("Unexpected EOF processing " + verbatimID);
 
-					tokenizer.Verbatim = false;
+						tokenizer.Verbatim = false;
 
-					EndElement ();
-					endPosition -= end_verbatim.Length;
-					OnTextParsed (verbatim_text);
-					beginPosition = endPosition;
-					endPosition += end_verbatim.Length;
-					OnTagParsed (TagType.Close, verbatimID, null);
-					continue;
-				}
-				
-				if (token == '<') {
-					GetTag (out tagtype, out id, out attributes);
-					EndElement ();
-					if (tagtype == TagType.ServerComment)
+						EndElement ();
+						endPosition -= end_verbatim.Length;
+						OnTextParsed (verbatim_text);
+						beginPosition = endPosition;
+						endPosition += end_verbatim.Length;
+						OnTagParsed (TagType.Close, verbatimID, null);
 						continue;
+					}
+				
+					if (token == '<') {
+						GetTag (out tagtype, out id, out attributes);
+						EndElement ();
+						if (tagtype == TagType.ServerComment)
+							continue;
 
-					if (tagtype == TagType.Text)
-						OnTextParsed (id);
-					else
-						OnTagParsed (tagtype, id, attributes);
+						if (tagtype == TagType.Text)
+							OnTextParsed (id);
+						else
+							OnTagParsed (tagtype, id, attributes);
 
-					continue;
+						continue;
+					}
+
+					if (tokenizer.Value.Trim ().Length == 0 && tagtype == TagType.Directive) {
+						continue;
+					}
+
+					text.Length = 0;
+					do {
+						text.Append (tokenizer.Value);
+						token = tokenizer.get_token ();
+					} while (token != '<' && token != Token.EOF);
+
+					tokenizer.put_back ();
+					EndElement ();
+					OnTextParsed (text.ToString ());
 				}
-
-				if (tokenizer.Value.Trim () == "" && tagtype == TagType.Directive) {
-					continue;
+			} finally {
+				if (fileReader != null) {
+					fileReader.Close ();
+					fileReader = null;
 				}
-
-				text.Length = 0;
-				do {
-					text.Append (tokenizer.Value);
-					token = tokenizer.get_token ();
-				} while (token != '<' && token != Token.EOF);
-
-				tokenizer.put_back ();
-				EndElement ();
-				OnTextParsed (text.ToString ());
+#if NET_2_0
+				checksum = tokenizer.Checksum;
+#endif
+				tokenizer = null;
 			}
 
 #if NET_2_0
@@ -358,9 +439,24 @@ namespace System.Web.Compilation
 
 				break;
 			default:
+				string idvalue = null;
+				// This is to handle code like:
+				//
+				//  <asp:ListItem runat="server"> < </asp:ListItem>
+				//
+				if ((char)token == '<') {
+					string odds = tokenizer.Odds;
+					if (odds != null && odds.Length > 0 && Char.IsWhiteSpace (odds [0])) {
+						tokenizer.put_back ();
+						idvalue = odds;
+					} else
+						idvalue = tokenizer.Value;
+				} else
+					idvalue = tokenizer.Value;
+				
 				tagtype = TagType.Text;
 				tokenizer.InTag = false;
-				id = "<" + tokenizer.Value;
+				id = "<" + idvalue;
 				break;
 			}
 		}
@@ -428,14 +524,14 @@ namespace System.Web.Compilation
 				token = tokenizer.get_token ();
 			}
 
-			end = end.ToLower (CultureInfo.InvariantCulture);
+			end = end.ToLower (Helpers.InvariantCulture);
 			int repeated = 0;
 			for (int k = 0; k < end.Length; k++)
 				if (end [0] == end [k])
 					repeated++;
 			
 			while (token != Token.EOF){
-				if (Char.ToLower ((char) token, CultureInfo.InvariantCulture) == end [i]){
+				if (Char.ToLower ((char) token, Helpers.InvariantCulture) == end [i]){
 					if (++i >= end.Length)
 						break;
 					tmp.Append ((char) token);
@@ -515,9 +611,18 @@ namespace System.Web.Compilation
 			bool databinding;
 			varname = Eat ('=');
 			databinding = !varname && Eat ('#');
-
+			string odds = tokenizer.Odds;
+			
 			tokenizer.Verbatim = true;
 			inside_tags = GetVerbatim (tokenizer.get_token (), "%>");
+			if (databinding && odds != null && odds.Length > 0) {
+				databinding = false;
+
+				// We encountered <% #something here %>, this should be passed
+				// verbatim to the compiler
+				inside_tags = '#' + inside_tags;
+			}			
+
 			tokenizer.Verbatim = false;
 			id = inside_tags;
 			attributes = null;
