@@ -65,15 +65,38 @@ namespace System.Web.UI {
 	[AspNetHostingPermission (SecurityAction.InheritanceDemand, Level = AspNetHostingPermissionLevel.Minimal)]
 	public abstract class TemplateParser : BaseParser
 	{
+#if NET_2_0
+		[Flags]
+		internal enum OutputCacheParsedParams
+		{
+			Location               = 0x0001,
+			CacheProfile           = 0x0002,
+			NoStore                = 0x0004,
+			SqlDependency          = 0x0008,
+			VaryByCustom           = 0x0010,
+			VaryByHeader           = 0x0020,
+			VaryByControl          = 0x0040,
+			VaryByContentEncodings = 0x0080
+		}
+#endif
+		
 		string inputFile;
 		string text;
 		Hashtable mainAttributes;
 		ArrayList dependencies;
 		ArrayList assemblies;
 		Hashtable anames;
-		ArrayList imports;
+#if NET_2_0
+		string[] binDirAssemblies;
+		Dictionary <string, bool> namespacesCache;
+		Dictionary <string, bool> imports;
+		List <string> interfaces;
+		List <ServerSideScript> scripts;
+#else
+		Hashtable imports;
 		ArrayList interfaces;
 		ArrayList scripts;
+#endif
 		Type baseType;
 		bool baseTypeIsGlobal = true;
 		string className;
@@ -89,12 +112,17 @@ namespace System.Web.UI {
 		int oc_duration;
 		string oc_header, oc_custom, oc_param, oc_controls;
 #if NET_2_0
-		string oc_content_encodings;
+		string oc_content_encodings, oc_cacheprofile, oc_sqldependency;
+		bool oc_nostore;
+		OutputCacheParsedParams oc_parsed_params = 0;
 #endif
 		bool oc_shared;
 		OutputCacheLocation oc_location;
-		CultureInfo invariantCulture = CultureInfo.InvariantCulture;
 #if NET_2_0
+		// Kludge needed to support pre-parsing of the main directive (see
+		// AspNetGenerator.GetRootBuilderType)
+		internal int allowedMainDirectives = 0;
+		
 		byte[] md5checksum;
 		string src;
 		bool srcIsLegacy;
@@ -102,7 +130,6 @@ namespace System.Web.UI {
 		string codeFileBaseClass;
 		string metaResourceKey;
 		Type codeFileBaseClassType;
-		string pageParserFilterTypeName;
 		Type pageParserFilterType;
 		PageParserFilter pageParserFilter;
 		
@@ -119,44 +146,32 @@ namespace System.Web.UI {
 
 		internal TemplateParser ()
 		{
-			LoadConfigDefaults ();
-			
-			imports = new ArrayList ();
 #if NET_2_0
-			AddNamespaces (imports);
+			imports = new Dictionary <string, bool> (StringComparer.Ordinal);
 #else
-			imports.Add ("System");
-			imports.Add ("System.Collections");
-			imports.Add ("System.Collections.Specialized");
-			imports.Add ("System.Configuration");
-			imports.Add ("System.Text");
-			imports.Add ("System.Text.RegularExpressions");
-			imports.Add ("System.Web");
-			imports.Add ("System.Web.Caching");
-			imports.Add ("System.Web.Security");
-			imports.Add ("System.Web.SessionState");
-			imports.Add ("System.Web.UI");
-			imports.Add ("System.Web.UI.WebControls");
-			imports.Add ("System.Web.UI.HtmlControls");
+			imports = new Hashtable ();
+			imports.Add ("System", true);
+			imports.Add ("System.Collections", true);
+			imports.Add ("System.Collections.Specialized", true);
+			imports.Add ("System.Configuration", true);
+			imports.Add ("System.Text", true);
+			imports.Add ("System.Text.RegularExpressions", true);
+			imports.Add ("System.Web", true);
+			imports.Add ("System.Web.Caching", true);
+			imports.Add ("System.Web.Security", true);
+			imports.Add ("System.Web.SessionState", true);
+			imports.Add ("System.Web.UI", true);
+			imports.Add ("System.Web.UI.WebControls", true);
+			imports.Add ("System.Web.UI.HtmlControls", true);
 #endif
-
+			LoadConfigDefaults ();
 			assemblies = new ArrayList ();
 #if NET_2_0
 			CompilationSection compConfig = CompilationConfig;
-			
-			bool addAssembliesInBin = false;
 			foreach (AssemblyInfo info in compConfig.Assemblies) {
-				if (info.Assembly == "*")
-					addAssembliesInBin = true;
-				else
+				if (info.Assembly != "*")
 					AddAssemblyByName (info.Assembly);
 			}
-			if (addAssembliesInBin)
-				AddAssembliesInBin ();
-
-			foreach (NamespaceInfo info in PagesConfig.Namespaces) {
-				imports.Add (info.Namespace);
-			}
 #else
 			CompilationConfiguration compConfig = CompilationConfig;
 			
@@ -172,10 +187,10 @@ namespace System.Web.UI {
 
 		internal virtual void LoadConfigDefaults ()
 		{
-			debug = CompilationConfig.Debug;
 #if NET_2_0
-			pageParserFilterTypeName = PagesConfig.PageParserFilterType;
+			AddNamespaces (imports);
 #endif
+			debug = CompilationConfig.Debug;
 		}
 		
 		internal void AddApplicationAssembly ()
@@ -200,21 +215,26 @@ namespace System.Web.UI {
 			generator.AddControl (type, attributes);
 		}
 		
-		void AddNamespaces (ArrayList imports)
+		void AddNamespaces (Dictionary <string, bool> imports)
 		{
 			if (BuildManager.HaveResources)
-				imports.Add ("System.Resources");
+				imports.Add ("System.Resources", true);
 			
-			PagesSection pages = WebConfigurationManager.GetWebApplicationSection ("system.web/pages") as PagesSection;
+			PagesSection pages = PagesConfig;
 			if (pages == null)
 				return;
 
 			NamespaceCollection namespaces = pages.Namespaces;
 			if (namespaces == null || namespaces.Count == 0)
 				return;
-
-			foreach (NamespaceInfo nsi in namespaces)
-				imports.Add (nsi.Namespace);
+			
+			foreach (NamespaceInfo nsi in namespaces) {
+				string ns = nsi.Namespace;
+				if (imports.ContainsKey (ns))
+					continue;
+				
+				imports.Add (ns, true);
+			}
 		}
 #endif
 		
@@ -242,7 +262,7 @@ namespace System.Web.UI {
 			if (!fileExists)
 				ThrowParseFileNotFound (src);
 
-			if (String.Compare (realpath, inputFile, false, invariantCulture) == 0)
+			if (String.Compare (realpath, inputFile, false, Helpers.InvariantCulture) == 0)
                                 return;
 			
 #if NET_2_0
@@ -301,19 +321,34 @@ namespace System.Web.UI {
 		{
 #if NET_2_0
 			var pageParserFilter = PageParserFilter;
-			if (pageParserFilter != null)
-				pageParserFilter.PreprocessDirective (directive.ToLower (CultureInfo.InvariantCulture), atts);
 #endif
-			if (String.Compare (directive, DefaultDirectiveName, true) == 0) {
-				if (mainAttributes != null)
+			if (String.Compare (directive, DefaultDirectiveName, true, Helpers.InvariantCulture) == 0) {
+#if NET_2_0
+				bool allowMainDirective = allowedMainDirectives > 0;
+#else
+				bool allowMainDirective = false;
+#endif
+				if (mainAttributes != null && !allowMainDirective)
 					ThrowParseException ("Only 1 " + DefaultDirectiveName + " is allowed");
-
+#if NET_2_0
+				allowedMainDirectives--;
+				if (mainAttributes != null)
+					return;
+				
+				if (pageParserFilter != null)
+					pageParserFilter.PreprocessDirective (directive.ToLower (Helpers.InvariantCulture), atts);
+#endif
+				
 				mainAttributes = atts;
 				ProcessMainAttributes (mainAttributes);
 				return;
 			}
-
-			int cmp = String.Compare ("Assembly", directive, true);
+#if NET_2_0
+			else if (pageParserFilter != null)
+				pageParserFilter.PreprocessDirective (directive.ToLower (Helpers.InvariantCulture), atts);
+#endif
+				
+			int cmp = String.Compare ("Assembly", directive, true, Helpers.InvariantCulture);
 			if (cmp == 0) {
 				string name = GetString (atts, "Name", null);
 				string src = GetString (atts, "Src", null);
@@ -336,18 +371,17 @@ namespace System.Web.UI {
 				return;
 			}
 
-			cmp = String.Compare ("Import", directive, true);
+			cmp = String.Compare ("Import", directive, true, Helpers.InvariantCulture);
 			if (cmp == 0) {
 				string namesp = GetString (atts, "Namespace", null);
 				if (atts.Count > 0)
 					ThrowParseException ("Attribute " + GetOneKey (atts) + " unknown.");
 				
-				if (namesp != null && namesp != "")
-					AddImport (namesp);
+				AddImport (namesp);
 				return;
 			}
 
-			cmp = String.Compare ("Implements", directive, true);
+			cmp = String.Compare ("Implements", directive, true, Helpers.InvariantCulture);
 			if (cmp == 0) {
 				string ifacename = GetString (atts, "Interface", "");
 
@@ -365,7 +399,7 @@ namespace System.Web.UI {
 				return;
 			}
 
-			cmp = String.Compare ("OutputCache", directive, true);
+			cmp = String.Compare ("OutputCache", directive, true, Helpers.InvariantCulture);
 			if (cmp == 0) {
 				HttpResponse response = HttpContext.Current.Response;
 				if (response != null)
@@ -382,7 +416,10 @@ namespace System.Web.UI {
 
 				foreach (DictionaryEntry entry in atts) {
 					string key = (string) entry.Key;
-					switch (key.ToLower ()) {
+					if (key == null)
+						continue;
+					
+					switch (key.ToLower (Helpers.InvariantCulture)) {
 						case "duration":
 							oc_duration = Int32.Parse ((string) entry.Value);
 							if (oc_duration < 1)
@@ -390,20 +427,46 @@ namespace System.Web.UI {
 										     "to a positive integer value");
 							break;
 #if NET_2_0
+						case "sqldependency":
+							oc_sqldependency = (string) entry.Value;
+							break;
+							
+						case "nostore":
+							try {
+								oc_nostore = Boolean.Parse ((string) entry.Value);
+								oc_parsed_params |= OutputCacheParsedParams.NoStore;
+							} catch {
+								ThrowParseException ("The 'NoStore' attribute is case sensitive" +
+										     " and must be set to 'true' or 'false'.");
+							}
+							break;
+
+						case "cacheprofile":
+							oc_cacheprofile = (string) entry.Value;
+							oc_parsed_params |= OutputCacheParsedParams.CacheProfile;
+							break;
+							
 						case "varybycontentencodings":
 							oc_content_encodings = (string) entry.Value;
+							oc_parsed_params |= OutputCacheParsedParams.VaryByContentEncodings;
 							break;
 #endif
 						case "varybyparam":
 							oc_param = (string) entry.Value;
-							if (String.Compare (oc_param, "none") == 0)
+							if (String.Compare (oc_param, "none", true, Helpers.InvariantCulture) == 0)
 								oc_param = null;
 							break;
 						case "varybyheader":
 							oc_header = (string) entry.Value;
+#if NET_2_0
+							oc_parsed_params |= OutputCacheParsedParams.VaryByHeader;
+#endif
 							break;
 						case "varybycustom":
 							oc_custom = (string) entry.Value;
+#if NET_2_0
+							oc_parsed_params |= OutputCacheParsedParams.VaryByCustom;
+#endif
 							break;
 						case "location":
 							if (!(this is PageParser))
@@ -412,6 +475,9 @@ namespace System.Web.UI {
 							try {
 								oc_location = (OutputCacheLocation) Enum.Parse (
 									typeof (OutputCacheLocation), (string) entry.Value, true);
+#if NET_2_0
+								oc_parsed_params |= OutputCacheParsedParams.Location;
+#endif
 							} catch {
 								ThrowParseException ("The 'location' attribute is case sensitive and " +
 										     "must be one of the following values: Any, Client, " +
@@ -424,6 +490,9 @@ namespace System.Web.UI {
 								goto default;
 #endif
 							oc_controls = (string) entry.Value;
+#if NET_2_0
+							oc_parsed_params |= OutputCacheParsedParams.VaryByControl;
+#endif
 							break;
 						case "shared":
 							if (this is PageParser)
@@ -471,14 +540,26 @@ namespace System.Web.UI {
 
 		void AddAssembliesInBin ()
 		{
-			foreach (string s in HttpApplication.BinDirectoryAssemblies)
-				assemblies.Add (s);
+			Assembly asm;
+			foreach (string s in HttpApplication.BinDirectoryAssemblies) {
+				try {
+					asm = Assembly.LoadFrom (s);
+					assemblies.Add (asm.Location);
+				} catch (BadImageFormatException) {
+					// ignore silently
+				}
+			}
 		}
 		
 		internal virtual void AddInterface (string iface)
 		{
-			if (interfaces == null)
+			if (interfaces == null) {
+#if NET_2_0
+				interfaces = new List <string> ();
+#else
 				interfaces = new ArrayList ();
+#endif
+			}
 
 			if (!interfaces.Contains (iface))
 				interfaces.Add (iface);
@@ -486,13 +567,82 @@ namespace System.Web.UI {
 		
 		internal virtual void AddImport (string namesp)
 		{
-			if (imports == null)
-				imports = new ArrayList ();
-
-			if (!imports.Contains (namesp))
-				imports.Add (namesp);
+			if (namesp == null || namesp.Length == 0)
+				return;
+			
+			if (imports == null) {
+#if NET_2_0
+				imports = new Dictionary <string, bool> (StringComparer.Ordinal);
+#else
+				imports = new Hashtable ();
+#endif
+			}
+			
+			if (imports.ContainsKey (namesp))
+				return;
+			
+			imports.Add (namesp, true);
+#if NET_2_0
+			AddAssemblyForNamespace (namesp);
+#endif
 		}
 
+#if NET_2_0
+		void AddAssemblyForNamespace (string namesp)
+		{
+			if (binDirAssemblies == null)
+				binDirAssemblies = HttpApplication.BinDirectoryAssemblies;
+			if (binDirAssemblies.Length == 0)
+				return;
+
+			if (namespacesCache == null)
+				namespacesCache = new Dictionary <string, bool> ();
+			else if (namespacesCache.ContainsKey (namesp))
+				return;
+			
+			foreach (Assembly asm in AppDomain.CurrentDomain.GetAssemblies ())
+				if (FindNamespaceInAssembly (asm, namesp))
+					return;
+			
+			IList tla = BuildManager.TopLevelAssemblies;
+			if (tla != null && tla.Count > 0) {
+				foreach (Assembly asm in tla) {
+					if (FindNamespaceInAssembly (asm, namesp))
+						return;
+				}
+			}
+
+			Assembly a;
+			foreach (string s in binDirAssemblies) {
+				a = Assembly.LoadFrom (s);
+				if (FindNamespaceInAssembly (a, namesp))
+					return;
+			}
+		}
+
+		bool FindNamespaceInAssembly (Assembly asm, string namesp)
+		{
+			Type[] asmTypes;
+
+			try {
+				asmTypes = asm.GetTypes ();
+			} catch (ReflectionTypeLoadException) {
+				// ignore
+				return false;
+			}
+			
+			foreach (Type type in asmTypes) {
+				if (String.Compare (type.Namespace, namesp, StringComparison.Ordinal) == 0) {
+					namespacesCache.Add (namesp, true);
+					AddAssembly (asm, true);
+					return true;
+				}
+			}
+
+			return false;
+		}
+#endif
+		
 		internal virtual void AddSourceDependency (string filename)
 		{
 			if (dependencies != null && dependencies.Contains (filename))
@@ -617,7 +767,7 @@ namespace System.Web.UI {
 			explicitOn = GetBool (atts, "Explicit", compConfig.Explicit);
 			if (atts.ContainsKey ("LinePragmas"))
 				linePragmasOn = GetBool (atts, "LinePragmas", true);
-			
+
 			string inherits = GetString (atts, "Inherits", null);
 #if NET_2_0
 			string srcRealPath = null;
@@ -754,7 +904,7 @@ namespace System.Web.UI {
 		{
 			MemberInfo mi = null;
 			bool missing = false;
-			string memberName = name.Trim ().ToLower (CultureInfo.InvariantCulture);
+			string memberName = name.Trim ().ToLower (Helpers.InvariantCulture);
 			Type parent = codeFileBaseClassType;
 
 			if (parent == null)
@@ -843,8 +993,11 @@ namespace System.Web.UI {
 					parent = LoadType (type);
 #endif				
 
-				if (parent == null)
+				if (parent == null) {
+					Console.WriteLine ("inputFile == {0}", inputFile);
+					Console.WriteLine (Environment.StackTrace);
 					ThrowParseException ("Cannot find type " + type);
+				}
 
 				if (!DefaultBaseType.IsAssignableFrom (parent))
 					ThrowParseException ("The parent type '" + type + "' does not derive from " + DefaultBaseType);
@@ -935,30 +1088,32 @@ namespace System.Web.UI {
 			set { md5checksum = value; }
 		}
 
-		internal string PageParserFilterTypeName {
-			get { return pageParserFilterTypeName; }
-		}
-
 		internal PageParserFilter PageParserFilter {
 			get {
 				if (pageParserFilter != null)
 					return pageParserFilter;
-				
-				if (String.IsNullOrEmpty (pageParserFilterTypeName))
-					return null;
 
-				pageParserFilter = Activator.CreateInstance (PageParserFilterType) as PageParserFilter;
-				pageParserFilter.Initialize (VirtualPath, this);
+				Type t = PageParserFilterType;
+				if (t == null)
+					return null;
 				
+				pageParserFilter = Activator.CreateInstance (t) as PageParserFilter;
+				pageParserFilter.Initialize (this);
+
 				return pageParserFilter;
 			}
 		}
 		
 		internal Type PageParserFilterType {
 			get {
-				if (pageParserFilterType == null)
-					pageParserFilterType = Type.GetType (PageParserFilterTypeName, true);
-
+				if (pageParserFilterType == null) {
+					string typeName = PagesConfig.PageParserFilterType;
+					if (String.IsNullOrEmpty (typeName))
+						return null;
+					
+					pageParserFilterType = HttpApplication.LoadType (typeName, true);
+				}
+				
 				return pageParserFilterType;
 			}
 		}
@@ -975,14 +1130,7 @@ namespace System.Web.UI {
 		internal ILocation DirectiveLocation {
 			get { return directiveLocation; }
 		}
-
-#if NET_2_0
-		internal VirtualPath VirtualPath {
-			get;
-			set;
-		}
-#endif
-
+		
 		internal string ParserDir {
 			get {
 				if (includeDirs == null || includeDirs.Count == 0)
@@ -1143,7 +1291,7 @@ namespace System.Web.UI {
 				}
 				
 				if (StrUtils.StartsWith (inFile, physPath))
-					className = inputFile.Substring (physPath.Length).ToLower (CultureInfo.InvariantCulture);
+					className = inputFile.Substring (physPath.Length).ToLower (Helpers.InvariantCulture);
 				else
 #endif
 					className = Path.GetFileName (inputFile);
@@ -1152,6 +1300,24 @@ namespace System.Web.UI {
 			}
 		}
 
+#if NET_2_0
+		internal List <ServerSideScript> Scripts {
+			get {
+				if (scripts == null)
+					scripts = new List <ServerSideScript> ();
+
+				return scripts;
+			}
+		}
+
+		internal Dictionary <string, bool> Imports {
+			get { return imports; }
+		}
+
+		internal List <string> Interfaces {
+			get { return interfaces; }
+		}
+#else
 		internal ArrayList Scripts {
 			get {
 				if (scripts == null)
@@ -1161,10 +1327,15 @@ namespace System.Web.UI {
 			}
 		}
 
-		internal ArrayList Imports {
+		internal Hashtable Imports {
 			get { return imports; }
 		}
 
+		internal ArrayList Interfaces {
+			get { return interfaces; }
+		}
+#endif
+		
 		internal ArrayList Assemblies {
 			get {
 				if (appAssemblyIndex != -1) {
@@ -1178,12 +1349,17 @@ namespace System.Web.UI {
 			}
 		}
 
-		internal ArrayList Interfaces {
-			get { return interfaces; }
-		}
-
 		internal RootBuilder RootBuilder {
-			get { return rootBuilder; }
+			get {
+#if NET_2_0
+				if (rootBuilder != null)
+					return rootBuilder;
+				AspGenerator generator = AspGenerator;
+				if (generator != null)
+					rootBuilder = generator.RootBuilder;
+#endif
+				return rootBuilder;
+			}
 			set { rootBuilder = value; }
 		}
 
@@ -1225,10 +1401,26 @@ namespace System.Web.UI {
 		}
 
 #if NET_2_0
+		internal OutputCacheParsedParams OutputCacheParsedParameters {
+			get { return oc_parsed_params; }
+		}
+
+		internal string OutputCacheSqlDependency {
+			get { return oc_sqldependency; }
+		}
+		
+		internal string OutputCacheCacheProfile {
+			get { return oc_cacheprofile; }
+		}
+		
 		internal string OutputCacheVaryByContentEncodings {
 			get { return oc_content_encodings; }
 		}
 
+		internal bool OutputCacheNoStore {
+			get { return oc_nostore; }
+		}
+		
 		internal virtual TextReader Reader {
 			get { return null; }
 			set { /* no-op */ }
@@ -1265,9 +1457,7 @@ namespace System.Web.UI {
 		}
 		
 		internal PagesSection PagesConfig {
-			get {
-				return WebConfigurationManager.GetWebApplicationSection ("system.web/pages") as PagesSection;
-			}
+			get { return GetConfigSection <PagesSection> ("system.web/pages") as PagesSection; }
 		}
 
 		internal AspGenerator AspGenerator {
@@ -1280,5 +1470,4 @@ namespace System.Web.UI {
 		}
 #endif
 	}
-}
-
+}
\ No newline at end of file
