@@ -50,7 +50,7 @@ namespace System.Xml.Schema
 		string stringValue;
 		XmlSchemaType schemaType;
 		XmlTypeCode xmlTypeCode;
-		ICollection valueAsList;
+		//ICollection valueAsList;
 
 		#region Constructors
 
@@ -305,6 +305,11 @@ namespace System.Xml.Schema
 			throw new NotImplementedException ();
 		}
 
+		public override string ToString ()
+		{
+			return Value;
+		}
+
 		#endregion
 
 		#region Properties
