@@ -153,8 +153,11 @@ namespace System.Xml.Schema
 		int xsiNilDepth = -1;
 		int skipValidationDepth = -1;
 
-		SOMObject currentType;
-
+		// LAMESPEC: XmlValueGetter is bogus by design because there
+		// is no way to get associated schema type for current value.
+		// Here XmlSchemaValidatingReader needs "current type"
+		// information to validate attribute values.
+		internal XmlSchemaDatatype CurrentAttributeType;
 
 		#endregion
 
@@ -333,14 +336,7 @@ namespace System.Xml.Schema
 		{
 			if (attributeValue == null)
 				throw new ArgumentNullException ("attributeValue");
-			return ValidateAttribute (localName, ns,
-				delegate () {
-					if (info != null && info.SchemaType != null && info.SchemaType.Datatype != null)
-						return info.SchemaType.Datatype.ParseValue (attributeValue, nameTable, nsResolver);
-					else
-						return attributeValue;
-				},
-				info);
+			return ValidateAttribute (localName, ns, delegate () { return attributeValue; }, info);
 		}
 
 		// I guess this weird XmlValueGetter is for such case that
@@ -502,7 +498,9 @@ namespace System.Xml.Schema
 				if (schemas.Count == 0)
 					return;
 
-				AssessCloseStartElementSchemaValidity (info);
+				if (skipValidationDepth < 0 || depth <= skipValidationDepth)
+					AssessCloseStartElementSchemaValidity (info);
+				depth++;
 			} finally {
 				occuredAtts.Clear ();
 			}
@@ -529,6 +527,9 @@ namespace System.Xml.Schema
 			if (schemas.Count == 0)
 				return;
 
+			if (skipValidationDepth >= 0 && depth > skipValidationDepth)
+				return;
+
 			ComplexType ct = Context.ActualType as ComplexType;
 			if (ct != null) {
 				switch (ct.ContentType) {
@@ -616,7 +617,7 @@ namespace System.Xml.Schema
 					throw new InvalidOperationException ("Initialize() must be called before processing validation.");
 				else
 					throw new InvalidOperationException (
-						String.Format ("Unexpected attempt to validation state transition from {0} to {1} was happened.",
+						String.Format ("Unexpected attempt to validate state transition from {0} to {1}.",
 							transition,
 							expected));
 			}
@@ -720,8 +721,6 @@ namespace System.Xml.Schema
 					next = state.Create (XmlSchemaParticle.Empty);
 			}
 			Context.State = next;
-
-			depth++;
 		}
 
 		// It must be invoked after xsi:nil turned out not to be in
@@ -822,6 +821,7 @@ namespace System.Xml.Schema
 			if (dt != SimpleType.AnySimpleType || attr.ValidatedFixedValue != null) {
 				object parsedValue = null;
 				try {
+					CurrentAttributeType = dt;
 					parsedValue = getter ();
 				} catch (Exception ex) { // It is inevitable and bad manner.
 					HandleError (String.Format ("Attribute value is invalid against its data type {0}", dt != null ? dt.TokenizedType : default (XmlTokenizedType)), ex);
@@ -832,11 +832,10 @@ namespace System.Xml.Schema
 				if (st != null)
 					ValidateRestrictedSimpleTypeValue (st, ref dt, new XmlAtomicValue (parsedValue, attr.AttributeSchemaType).Value);
 
-				XmlSchemaType type = info != null ? info.SchemaType : null;
 				if (attr.ValidatedFixedValue != null && 
-					!XmlSchemaUtil.IsSchemaDatatypeEquals (
-					attr.AttributeSchemaType.Datatype as XsdAnySimpleType, attr.ValidatedFixedTypedValue, type != null ? type.Datatype as XsdAnySimpleType : null, parsedValue)) {
-					HandleError ("The value of the attribute " + attr.QualifiedName + " does not match with its fixed value.");
+					!XmlSchemaUtil.AreSchemaDatatypeEqual (
+					attr.AttributeSchemaType.Datatype as XsdAnySimpleType, attr.ValidatedFixedTypedValue, dt as XsdAnySimpleType, parsedValue)) {
+					HandleError (String.Format ("The value of the attribute {0} does not match with its fixed value '{1}' in the space of type {2}", attr.QualifiedName, attr.ValidatedFixedValue, dt));
 					parsedValue = attr.ValidatedFixedTypedValue;
 				}
 #region ID Constraints
@@ -909,8 +908,10 @@ namespace System.Xml.Schema
 			if (xsiNilDepth >= 0 && xsiNilDepth < depth)
 				HandleError ("Element item appeared, while current element context is nil.");
 
-			if (shouldValidateCharacters)
+			if (shouldValidateCharacters) {
+				CurrentAttributeType = null;
 				storedCharacters.Append (getter ());
+			}
 		}
 
 
@@ -1229,6 +1230,7 @@ namespace System.Xml.Schema
 				XsdKeyTable seq  = (XsdKeyTable) keyTables [i];
 				// If possible, create new field entry candidates.
 				for (int j = 0; j < seq.Entries.Count; j++) {
+					CurrentAttributeType = null;
 					try {
 						seq.Entries [j].ProcessMatch (
 							isAttr,
@@ -1254,10 +1256,6 @@ namespace System.Xml.Schema
 			}
 		}
 
-		private void ProcessKeyEntryOne (XsdKeyEntry entry, bool isAttr, bool isNil, object schemaType, string attrName, string attrNs, XmlValueGetter getter)
-		{
-		}
-
 		private void ValidateEndElementKeyConstraints ()
 		{
 			// Reset Identity constraints.
