@@ -35,6 +35,9 @@ using System.Runtime.Serialization;
 using System.Globalization;
 #if NET_2_0
 using System.Net.Configuration;
+using System.Net.Security;
+using System.Net.Cache;
+using System.Security.Principal;
 #endif
 
 namespace System.Net 
@@ -43,6 +46,10 @@ namespace System.Net
 	public abstract class WebRequest : MarshalByRefObject, ISerializable
 	{
 		static HybridDictionary prefixes = new HybridDictionary ();
+#if NET_2_0
+		static bool isDefaultWebProxySet;
+		static IWebProxy defaultWebProxy;
+#endif
 		
 		// Constructors
 		
@@ -67,6 +74,9 @@ namespace System.Net
 		
 		protected WebRequest (SerializationInfo serializationInfo, StreamingContext streamingContext) 
 		{
+#if ONLY_1_1
+			throw GetMustImplement ();
+#endif
 		}
 
 		static Exception GetMustImplement ()
@@ -75,8 +85,31 @@ namespace System.Net
 		}
 		
 		// Properties
+
+#if NET_2_0
+		private AuthenticationLevel authentication_level = AuthenticationLevel.MutualAuthRequested;
 		
-		public virtual string ConnectionGroupName { 
+		public AuthenticationLevel AuthenticationLevel
+		{
+			get {
+				return(authentication_level);
+			}
+			set {
+				authentication_level = value;
+			}
+		}
+
+		public virtual RequestCachePolicy CachePolicy
+		{
+			get {
+				throw GetMustImplement ();
+			}
+			set {
+			}
+		}
+#endif
+		
+		public virtual string ConnectionGroupName {
 			get { throw GetMustImplement (); }
 			set { throw GetMustImplement (); }
 		}
@@ -95,12 +128,30 @@ namespace System.Net
 			get { throw GetMustImplement (); }
 			set { throw GetMustImplement (); }
 		}
+
+#if NET_2_0
+		public static RequestCachePolicy DefaultCachePolicy
+		{
+			get {
+				throw GetMustImplement ();
+			}
+			set {
+				throw GetMustImplement ();
+			}
+		}
+#endif
 		
 		public virtual WebHeaderCollection Headers { 
 			get { throw GetMustImplement (); }
 			set { throw GetMustImplement (); }
 		}
 		
+#if NET_2_0
+		public TokenImpersonationLevel ImpersonationLevel {
+			get { throw GetMustImplement (); }
+			set { throw GetMustImplement (); }
+		}
+#endif
 		public virtual string Method { 
 			get { throw GetMustImplement (); }
 			set { throw GetMustImplement (); }
@@ -126,22 +177,36 @@ namespace System.Net
 		}
 		
 #if NET_2_0
-		volatile static IWebProxy proxy;
+		public virtual bool UseDefaultCredentials
+		{
+			get {
+				throw GetMustImplement ();
+			}
+			set {
+				throw GetMustImplement ();
+			}
+		}
+		
+//		volatile static IWebProxy proxy;
 		static readonly object lockobj = new object ();
 		
 		public static IWebProxy DefaultWebProxy {
 			get {
-				lock (lockobj) {
-					if (proxy == null)
-						proxy = GetDefaultWebProxy ();
-					return proxy;
+				if (!isDefaultWebProxySet) {
+					lock (lockobj) {
+						if (defaultWebProxy == null)
+							defaultWebProxy = GetDefaultWebProxy ();
+					}
 				}
+				return defaultWebProxy;
 			}
 			set {
-				if (value == null)
-					throw new ArgumentNullException ("WebRequest.DefaultWebProxy",
-							"null IWebProxy not allowed.");
-				proxy = value;
+				/* MS documentation states that a null value would cause an ArgumentNullException
+				 * but that's not the way it behaves:
+				 * https://connect.microsoft.com/VisualStudio/feedback/ViewFeedback.aspx?FeedbackID=304724
+				 */
+				defaultWebProxy = value;
+				isDefaultWebProxySet = true;
 			}
 		}
 		
@@ -169,7 +234,7 @@ namespace System.Net
 				p.BypassProxyOnLocal = (pe.BypassOnLocal == ProxyElement.BypassOnLocalValues.True);
 #endif
 			return p;
-		}		
+		}
 #endif
 
 		// Methods
@@ -235,32 +300,54 @@ namespace System.Net
 		public static IWebProxy GetSystemWebProxy ()
 		{
 			string address = Environment.GetEnvironmentVariable ("http_proxy");
+			if (address == null)
+				address = Environment.GetEnvironmentVariable ("HTTP_PROXY");
+
 			if (address != null) {
 				try {
-					WebProxy p = new WebProxy (address);
-					return p;
-				} catch (UriFormatException) {}
+					if (!address.StartsWith ("http://"))
+						address = "http://" + address;
+					Uri uri = new Uri (address);
+					IPAddress ip;
+					if (IPAddress.TryParse (uri.Host, out ip)) {
+						if (IPAddress.Any.Equals (ip)) {
+							UriBuilder builder = new UriBuilder (uri);
+							builder.Host = "127.0.0.1";
+							uri = builder.Uri;
+						} else if (IPAddress.IPv6Any.Equals (ip)) {
+							UriBuilder builder = new UriBuilder (uri);
+							builder.Host = "[::1]";
+							uri = builder.Uri;
+						}
+					}
+					return new WebProxy (uri);
+				} catch (UriFormatException) { }
 			}
 			return new WebProxy ();
 		}
 #endif
-#if TARGET_JVM
-		public virtual void GetObjectData
-#else
+
 		void ISerializable.GetObjectData
-#endif
 		(SerializationInfo serializationInfo,
 		   				  StreamingContext streamingContext)
 		{
 			throw new NotSupportedException ();
 		}
 
+
+#if NET_2_0
+		protected virtual void GetObjectData (SerializationInfo serializationInfo, StreamingContext streamingContext)
+		{
+			throw GetMustImplement ();
+		}
+#endif
+
 		public static bool RegisterPrefix (string prefix, IWebRequestCreate creator)
 		{
 			if (prefix == null)
-				throw new ArgumentNullException("prefix");
+				throw new ArgumentNullException ("prefix");
 			if (creator == null)
-				throw new ArgumentNullException("creator");			
+				throw new ArgumentNullException ("creator");
 			
 			lock (prefixes.SyncRoot) {
 				string lowerCasePrefix = prefix.ToLower (CultureInfo.InvariantCulture);
@@ -286,7 +373,7 @@ namespace System.Net
 					continue;
 				
 				if (!prefix.StartsWith (key))
-					continue;					
+					continue;
 					
 				longestPrefix = key.Length;
 				creator = (IWebRequestCreate) e.Value;
