@@ -1,4 +1,4 @@
-// Calendar.cs
+// System.Globalization.Calendar.cs
 //
 // (C) Ulrich Kunitz 2002
 //
@@ -30,11 +30,15 @@ namespace System.Globalization {
 
 using System;
 using System.IO;
+using System.Runtime.InteropServices;
 
 /// <remarks>
 /// The class serves as a base class for calendar classes.
 /// </remarks>
 [Serializable]
+#if NET_2_0
+[ComVisible (true)]
+#endif
 public abstract class Calendar
 #if NET_2_0
 	: ICloneable
@@ -152,24 +156,43 @@ public abstract class Calendar
 	}
 
 	// LAMESPEC: huh, why not Calendar but Object?
+	[ComVisible (false)]
 	public virtual object Clone ()
 	{
 		Calendar c = (Calendar) MemberwiseClone ();
 		c.m_isReadOnly = false;
 		return c;
 	}
+
+	[ComVisible (false)]
+	public virtual int GetLeapMonth (int year)
+	{
+		return GetLeapMonth (year, GetEra (ToDateTime (year, 1, 1, 0, 0, 0, 0)));
+	}
+
+	[ComVisible (false)]
+	public virtual int GetLeapMonth (int year, int era)
+	{
+		int max = GetMonthsInYear (year, era);
+		for (int i = 1; i <= max; i++)
+			if (IsLeapMonth (year, i, era))
+				return i;
+		return 0;
+	}
 #endif
 
 #if NET_2_0
+	[ComVisible (false)]
 	public bool IsReadOnly {
 		get { return m_isReadOnly; }
 	}
 
-	public static Calendar ReadOnly (Calendar source)
+	[ComVisible (false)]
+	public static Calendar ReadOnly (Calendar calendar)
 	{
-		if (source.m_isReadOnly)
-			return source;
-		Calendar c = (Calendar) source.Clone ();
+		if (calendar.m_isReadOnly)
+			return calendar;
+		Calendar c = (Calendar) calendar.Clone ();
 		c.m_isReadOnly = true;
 		return c;
 	}
@@ -867,10 +890,10 @@ public abstract class Calendar
 	/// is out of range.
 	/// </exception>
 	public virtual DateTime ToDateTime(int year, int month, int day,
-		int hour, int minute, int second, int milliseconds)
+		int hour, int minute, int second, int millisecond)
 	{
-		return ToDateTime(year, month, day, hour, minute, second, 
-			milliseconds, CurrentEra);
+		return ToDateTime (year, month, day, hour, minute, second, 
+			millisecond, CurrentEra);
 	}
 
 
@@ -904,7 +927,7 @@ public abstract class Calendar
 	/// is out of range.
 	/// </exception>
 	public abstract DateTime ToDateTime(int year, int month, int day,
-		int hour, int minute, int second, int milliseconds,
+		int hour, int minute, int second, int millisecond,
 		int era);
 
 	/// <summary>
