@@ -169,6 +169,7 @@ namespace System.Runtime.Remoting.Proxies
 			CallType call_type = mMsg.CallType;
 			bool is_remproxy = (rp is RemotingProxy);
 
+			out_args = null;
 			IMethodReturnMessage res_msg = null;
 			
 			if (call_type == CallType.BeginInvoke) 
@@ -214,6 +215,7 @@ namespace System.Runtime.Remoting.Proxies
 
 					// allow calltype EndInvoke to finish
 					asyncMsg = mMsg.AsyncResult.SyncProcessMessage (res_msg as IMessage);
+					out_args = res_msg.OutArgs;
 					res_msg = new ReturnMessage (asyncMsg, null, 0, null, res_msg as IMethodCallMessage);
 				}
 			}
@@ -229,7 +231,8 @@ namespace System.Runtime.Remoting.Proxies
 				throw exc.FixRemotingException();
 			}
 			else if (res_msg is IConstructionReturnMessage || mMsg.CallType == CallType.BeginInvoke) {
-				out_args = res_msg.OutArgs;
+				if (out_args == null)
+					out_args = res_msg.OutArgs;
 			}
 			else if (mMsg.CallType == CallType.Sync) {
 				out_args = ProcessResponse (res_msg, mMsg);
@@ -238,7 +241,8 @@ namespace System.Runtime.Remoting.Proxies
 				out_args = ProcessResponse (res_msg, mMsg.AsyncResult.CallMessage);
 			}
 			else {
-				out_args = res_msg.OutArgs;
+				if (out_args == null)
+					out_args = res_msg.OutArgs;
 			}
 
 			return res_msg.ReturnValue;
