@@ -38,9 +38,13 @@ using System.Runtime.CompilerServices;
 using System.Security;
 using System.Security.Permissions;
 using System.Text;
+using System.Runtime.InteropServices;
 
 namespace System {
 
+#if NET_2_0
+	[ComVisible (true)]
+#endif
 #if NET_2_0
 	public static class Environment {
 #else
@@ -59,8 +63,11 @@ namespace System {
 		 * Changes which are already detected at runtime, like the addition
 		 * of icalls, do not require an increment.
 		 */
-		private const int mono_corlib_version = 56;
-		
+		private const int mono_corlib_version = 69;
+
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public enum SpecialFolder
 		{	// TODO: Determine if these windoze style folder identifiers 
 			//       have unix/linux counterparts
@@ -180,10 +187,6 @@ namespace System {
 				if (os == null) {
 					Version v = Version.CreateFromString (GetOSVersionString ());
 					PlatformID p = Platform;
-#if NET_2_0
-					if ((int) p == 128)
-						p = PlatformID.Unix;
-#endif
 					os = new OperatingSystem (p, v);
 				}
 				return os;
@@ -196,7 +199,7 @@ namespace System {
 		public static string StackTrace {
 			[EnvironmentPermission (SecurityAction.Demand, Unrestricted=true)]
 			get {
-				System.Diagnostics.StackTrace trace = new System.Diagnostics.StackTrace (1, true);
+				System.Diagnostics.StackTrace trace = new System.Diagnostics.StackTrace (0, true);
 				return trace.ToString ();
 			}
 		}
@@ -344,18 +347,18 @@ namespace System {
 		public extern static string[] GetCommandLineArgs ();
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		internal extern static string internalGetEnvironmentVariable (string name);
+		internal extern static string internalGetEnvironmentVariable (string variable);
 
 		/// <summary>
 		/// Return a string containing the value of the environment
 		/// variable identifed by parameter "variable"
 		/// </summary>
-		public static string GetEnvironmentVariable (string name)
+		public static string GetEnvironmentVariable (string variable)
 		{
 			if (SecurityManager.SecurityEnabled) {
-				new EnvironmentPermission (EnvironmentPermissionAccess.Read, name).Demand ();
+				new EnvironmentPermission (EnvironmentPermissionAccess.Read, variable).Demand ();
 			}
-			return internalGetEnvironmentVariable (name);
+			return internalGetEnvironmentVariable (variable);
 		}
 
 		static Hashtable GetEnvironmentVariablesNoCase ()
@@ -433,6 +436,48 @@ namespace System {
 			return dir;
 		}
 
+		private static string ReadXdgUserDir (string config_dir, string home_dir, 
+			string key, string fallback)
+		{
+			string env_path = internalGetEnvironmentVariable (key);
+			if (env_path != null && env_path != String.Empty) {
+				return env_path;
+			}
+
+			string user_dirs_path = Path.Combine (config_dir, "user-dirs.dirs");
+
+			if (!File.Exists (user_dirs_path)) {
+				return Path.Combine (home_dir, fallback);
+			}
+
+			try {
+				using(StreamReader reader = new StreamReader (user_dirs_path)) {
+					string line;
+					while ((line = reader.ReadLine ()) != null) {
+						line = line.Trim ();
+						int delim_index = line.IndexOf ('=');
+                        if(delim_index > 8 && line.Substring (0, delim_index) == key) {
+                            string path = line.Substring (delim_index + 1).Trim ('"');
+                            bool relative = false;
+
+                            if (path.StartsWith ("$HOME/")) {
+                                relative = true;
+                                path = path.Substring (6);
+                            } else if (!path.StartsWith ("/")) {
+                                relative = true;
+                            }
+
+                            return relative ? Path.Combine (home_dir, path) : path;
+                        }
+					}
+				}
+			} catch (FileNotFoundException) {
+			}
+
+			return Path.Combine (home_dir, fallback);
+		}
+
+
 		// the security runtime (and maybe other parts of corlib) needs the
 		// information to initialize themselves before permissions can be checked
 		internal static string InternalGetFolderPath (SpecialFolder folder)
@@ -473,10 +518,13 @@ namespace System {
 			case SpecialFolder.Desktop:
 #endif
 			case SpecialFolder.DesktopDirectory:
-				return Path.Combine (home, "Desktop");
+				return ReadXdgUserDir (config, home, "XDG_DESKTOP_DIR", "Desktop");
 
 			case SpecialFolder.MyMusic:
-				return Path.Combine (home, "Music");
+				return ReadXdgUserDir (config, home, "XDG_MUSIC_DIR", "Music");
+
+			case SpecialFolder.MyPictures:
+				return ReadXdgUserDir (config, home, "XDG_PICTURES_DIR", "Pictures");
 				
 			// these simply dont exist on Linux
 			// The spec says if a folder doesnt exist, we
@@ -486,7 +534,6 @@ namespace System {
 			case SpecialFolder.SendTo:
 			case SpecialFolder.StartMenu:
 			case SpecialFolder.Startup:
-			case SpecialFolder.MyPictures:
 			case SpecialFolder.Templates:
 			case SpecialFolder.Cookies:
 			case SpecialFolder.History:
@@ -510,10 +557,9 @@ namespace System {
 			return GetLogicalDrivesInternal ();
 		}
 
-		// FIXME: Anyone using this anywhere ?
-		static internal string GetResourceString (string s) { return String.Empty; }
+		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+		private static extern void internalBroadcastSettingChange ();
 
-                
 #if NET_2_0
 		public static string GetEnvironmentVariable (string variable, EnvironmentVariableTarget target)
 		{
@@ -525,14 +571,16 @@ namespace System {
 				if (!IsRunningOnWindows)
 					return null;
 				using (Microsoft.Win32.RegistryKey env = Microsoft.Win32.Registry.LocalMachine.OpenSubKey (@"SYSTEM\CurrentControlSet\Control\Session Manager\Environment")) {
-					return env.GetValue (variable).ToString ();
+					object regvalue = env.GetValue (variable);
+					return (regvalue == null) ? null : regvalue.ToString ();
 				}
 			case EnvironmentVariableTarget.User:
 				new EnvironmentPermission (PermissionState.Unrestricted).Demand ();
 				if (!IsRunningOnWindows)
 					return null;
 				using (Microsoft.Win32.RegistryKey env = Microsoft.Win32.Registry.CurrentUser.OpenSubKey ("Environment", false)) {
-					return env.GetValue (variable).ToString ();
+					object regvalue = env.GetValue (variable);
+					return (regvalue == null) ? null : regvalue.ToString ();
 				}
 			default:
 				throw new ArgumentException ("target");
@@ -598,20 +646,22 @@ namespace System {
 				if (!IsRunningOnWindows)
 					return;
 				using (Microsoft.Win32.RegistryKey env = Microsoft.Win32.Registry.LocalMachine.OpenSubKey (@"SYSTEM\CurrentControlSet\Control\Session Manager\Environment", true)) {
-					if (value == null || value.Length == 0)
+					if (String.IsNullOrEmpty (value))
 						env.DeleteValue (variable, false);
 					else
 						env.SetValue (variable, value);
+					internalBroadcastSettingChange ();
 				}
 				break;
 			case EnvironmentVariableTarget.User:
 				if (!IsRunningOnWindows)
 					return;
 				using (Microsoft.Win32.RegistryKey env = Microsoft.Win32.Registry.CurrentUser.OpenSubKey ("Environment", true)) {
-					if (value == null || value.Length == 0)
+					if (String.IsNullOrEmpty (value))
 						env.DeleteValue (variable, false);
 					else
 						env.SetValue (variable, value);
+					internalBroadcastSettingChange ();
 				}
 				break;
 			default:
@@ -639,9 +689,13 @@ namespace System {
 		// private methods
 
 		internal static bool IsRunningOnWindows {
-			get { return ((int) Platform != 128); }
+			get { return ((int) Platform < 4); }
 		}
 
+		//
+		// Used by gacutil.exe
+		//
+#pragma warning disable 169		
 		private static string GacPath {
 			get {
 				if (Environment.IsRunningOnWindows) {
@@ -653,6 +707,7 @@ namespace System {
 				return Path.Combine (Path.Combine (internalGetGacPath (), "mono"), "gac");
 			}
 		}
+#pragma warning restore 169
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern static string [] GetLogicalDrivesInternal ();
