@@ -229,6 +229,10 @@ namespace System {
 							"are permitted.");
 					return false;
 				}
+			} else if ((uint) style > (uint) NumberStyles.Any){
+				if (!tryParse)
+					exc = new ArgumentException ("Not a valid number style");
+				return false;
 			}
 
 			return true;
@@ -327,12 +331,6 @@ namespace System {
 			result = 0;
 			exc = null;
 
-			if (s == null) {
-				if (!tryParse)
-					exc = GetFormatException ();
-				return false;
-			}
-			
 			if (s == null) {
 				if (!tryParse)
 					exc = new ArgumentNullException ();
@@ -345,12 +343,12 @@ namespace System {
 				return false;
 			}
 
-			NumberFormatInfo nfi;
+			NumberFormatInfo nfi = null;
 			if (fp != null) {
 				Type typeNFI = typeof (System.Globalization.NumberFormatInfo);
 				nfi = (NumberFormatInfo) fp.GetFormat (typeNFI);
 			}
-			else
+			if (nfi == null)
 				nfi = Thread.CurrentThread.CurrentCulture.NumberFormat;
 
 			if (!CheckStyle (style, tryParse, ref exc))
