@@ -980,7 +980,10 @@ namespace System
 
 		public int LastIndexOf (string value, StringComparison comparisonType)
 		{
-			return LastIndexOf (value, this.Length - 1, this.Length, comparisonType);
+			if (this.Length == 0)
+				return value == String.Empty ? 0 : -1;
+			else
+				return LastIndexOf (value, this.Length - 1, this.Length, comparisonType);
 		}
 
 		public int LastIndexOf (string value, int startIndex, StringComparison comparisonType)
@@ -1180,7 +1183,7 @@ namespace System
 				return 0;
 			if (this.length == 0)
 				return -1;
-			return CultureInfo.CurrentCulture.CompareInfo.IndexOf (this, value, 0, length);
+			return CultureInfo.CurrentCulture.CompareInfo.IndexOf (this, value, 0, length, CompareOptions.Ordinal);
 		}
 
 		public int IndexOf (String value, int startIndex)
@@ -2424,7 +2427,7 @@ namespace System
 
 		object IConvertible.ToType (Type type, IFormatProvider provider)
 		{
-			return Convert.ToType (this, type, provider);
+			return Convert.ToType (this, type, provider, false);
 		}
 
 #if ONLY_1_1
