@@ -55,7 +55,7 @@ namespace System {
 		public static readonly MemberFilter FilterAttribute = new MemberFilter (FilterAttribute_impl);
 		public static readonly MemberFilter FilterName = new MemberFilter (FilterName_impl);
 		public static readonly MemberFilter FilterNameIgnoreCase = new MemberFilter (FilterNameIgnoreCase_impl);
-		public static readonly object Missing;
+		public static readonly object Missing = System.Reflection.Missing.Value;
 
 		internal const BindingFlags DefaultBindingFlags =
 		BindingFlags.Public | BindingFlags.Static | BindingFlags.Instance;
@@ -655,6 +655,8 @@ namespace System {
 				throw new ArgumentNullException ("interfaceType");
 			if (!interfaceType.IsInterface)
 				throw new ArgumentException (Locale.GetText ("Argument must be an interface."), "interfaceType");
+			if (IsInterface)
+				throw new ArgumentException ("'this' type cannot be an interface itself");
 			res.TargetType = this;
 			res.InterfaceType = interfaceType;
 			GetInterfaceMapData (this, interfaceType, out res.TargetMethods, out res.InterfaceMethods);
