@@ -129,6 +129,8 @@ static GSList *loaded_assembly_bindings = NULL;
 
 static MonoAssembly*
 mono_assembly_invoke_search_hook_internal (MonoAssemblyName *aname, gboolean refonly, gboolean postload);
+static MonoBoolean
+mono_assembly_is_in_gac (const gchar *filanem);
 
 static gchar*
 encode_public_tok (const guchar *token, gint32 len)
@@ -847,7 +849,7 @@ mono_assembly_load_reference (MonoImage *image, int index)
 		reference = mono_assembly_load (&aname, image->assembly? image->assembly->basedir: NULL, &status);
 
 	if (reference == NULL){
-		char *extra_msg = g_strdup ("");
+		char *extra_msg;
 
 		if (status == MONO_IMAGE_ERROR_ERRNO && errno == ENOENT) {
 			extra_msg = g_strdup_printf ("The assembly was not found in the Global Assembly Cache, a path listed in the MONO_PATH environment variable, or in the location of the executing assembly (%s).\n", image->assembly != NULL ? image->assembly->basedir : "" );
@@ -857,6 +859,8 @@ mono_assembly_load_reference (MonoImage *image, int index)
 			extra_msg = g_strdup ("Cannot find an assembly referenced from this one.\n");
 		} else if (status == MONO_IMAGE_IMAGE_INVALID) {
 			extra_msg = g_strdup ("The file exists but is not a valid assembly.\n");
+		} else {
+			extra_msg = g_strdup ("");
 		}
 		
 		g_warning ("The following assembly referenced from %s could not be loaded:\n"
@@ -1264,7 +1268,10 @@ mono_assembly_open_full (const char *filename, MonoImageOpenStatus *status, gboo
 
 	mono_trace (G_LOG_LEVEL_INFO, MONO_TRACE_ASSEMBLY,
 			"Assembly Loader probing location: '%s'.", fname);
-	new_fname = mono_make_shadow_copy (fname);
+
+	new_fname = NULL;
+	if (!mono_assembly_is_in_gac (fname))
+		new_fname = mono_make_shadow_copy (fname);
 	if (new_fname && new_fname != fname) {
 		g_free (fname);
 		fname = new_fname;
@@ -1602,6 +1609,10 @@ parse_public_key (const gchar *key, gchar** pubkey)
 	bitlen = read32 (header + 12) >> 3;
 	if ((bitlen + 16 + 4) != pkeylen)
 		return FALSE;
+
+	/* parsing is OK and the public key itself is not requested back */
+	if (!pubkey)
+		return TRUE;
 		
 	/* Encode the size of the blob */
 	offset = 0;
@@ -1618,8 +1629,8 @@ parse_public_key (const gchar *key, gchar** pubkey)
 		arr [i] = g_ascii_xdigit_value (key [j++]) << 4;
 		arr [i] |= g_ascii_xdigit_value (key [j++]);
 	}
-	if (pubkey)
-		*pubkey = arr;
+
+	*pubkey = arr;
 
 	return TRUE;
 }
@@ -1988,6 +1999,67 @@ mono_assembly_load_with_partial_name (const char *name, MonoImageOpenStatus *sta
 	return res;
 }
 
+static MonoBoolean
+mono_assembly_is_in_gac (const gchar *filename)
+{
+	const gchar *rootdir;
+	gchar *gp;
+	gchar **paths;
+
+	if (filename == NULL)
+		return FALSE;
+
+	for (paths = extra_gac_paths; paths && *paths; paths++) {
+		if (strstr (*paths, filename) != *paths)
+			continue;
+
+		gp = (gchar *) (filename + strlen (*paths));
+		if (*gp != G_DIR_SEPARATOR)
+			continue;
+		gp++;
+		if (strncmp (gp, "lib", 3))
+			continue;
+		gp += 3;
+		if (*gp != G_DIR_SEPARATOR)
+			continue;
+		gp++;
+		if (strncmp (gp, "mono", 4))
+			continue;
+		gp += 4;
+		if (*gp != G_DIR_SEPARATOR)
+			continue;
+		gp++;
+		if (strncmp (gp, "gac", 3))
+			continue;
+		gp += 3;
+		if (*gp != G_DIR_SEPARATOR)
+			continue;
+
+		return TRUE;
+	}
+
+	rootdir = mono_assembly_getrootdir ();
+	if (strstr (filename, rootdir) != filename)
+		return FALSE;
+
+	gp = (gchar *) (filename + strlen (rootdir));
+	if (*gp != G_DIR_SEPARATOR)
+		return FALSE;
+	gp++;
+	if (strncmp (gp, "mono", 4))
+		return FALSE;
+	gp += 4;
+	if (*gp != G_DIR_SEPARATOR)
+		return FALSE;
+	gp++;
+	if (strncmp (gp, "gac", 3))
+		return FALSE;
+	gp += 3;
+	if (*gp != G_DIR_SEPARATOR)
+		return FALSE;
+	return TRUE;
+}
+
 static MonoImage*
 mono_assembly_load_publisher_policy (MonoAssemblyName *aname)
 {
