@@ -20,9 +20,11 @@
 #include <mono/metadata/exception.h>
 #include <mono/metadata/gc-internal.h>
 #include <mono/metadata/mono-debug.h>
+#include <mono/utils/mono-mmap.h>
 
 #include "mini.h"
 #include "mini-x86.h"
+#include "debug-mini.h"
 
 #ifdef PLATFORM_WIN32
 static void (*restore_stack) (void *);
@@ -129,7 +131,7 @@ win32_handle_stack_overflow (EXCEPTION_POINTERS* ep, struct sigcontext *sctx)
 	do {
 		MonoContext new_ctx;
 
-		ji = mono_arch_find_jit_info (domain, jit_tls, &rji, &rji, &ctx, &new_ctx, NULL, &lmf, NULL, NULL);
+		ji = mono_arch_find_jit_info (domain, jit_tls, &rji, &rji, &ctx, &new_ctx, &lmf, NULL);
 		if (!ji) {
 			g_warning ("Exception inside function without unwind info");
 			g_assert_not_reached ();
@@ -333,9 +335,19 @@ mono_arch_get_call_filter (void)
 	x86_mov_reg_membase (code, X86_ESI, X86_EAX,  G_STRUCT_OFFSET (MonoContext, esi), 4);
 	x86_mov_reg_membase (code, X86_EDI, X86_EAX,  G_STRUCT_OFFSET (MonoContext, edi), 4);
 
+	/* align stack and save ESP */
+	x86_mov_reg_reg (code, X86_EDX, X86_ESP, 4);
+	x86_alu_reg_imm (code, X86_AND, X86_ESP, -MONO_ARCH_FRAME_ALIGNMENT);
+	g_assert (MONO_ARCH_FRAME_ALIGNMENT >= 8);
+	x86_alu_reg_imm (code, X86_SUB, X86_ESP, MONO_ARCH_FRAME_ALIGNMENT - 8);
+	x86_push_reg (code, X86_EDX);
+
 	/* call the handler */
 	x86_call_reg (code, X86_ECX);
 
+	/* restore ESP */
+	x86_pop_reg (code, X86_ESP);
+
 	/* restore EBP */
 	x86_pop_reg (code, X86_EBP);
 
@@ -362,7 +374,7 @@ throw_exception (unsigned long eax, unsigned long ecx, unsigned long edx, unsign
 		restore_context = mono_arch_get_restore_context ();
 
 	/* Pop argument and return address */
-	ctx.esp = esp + (2 * sizeof (gpointer));
+	ctx.esp = esp + (3 * sizeof (gpointer));
 	ctx.eip = eip;
 	ctx.ebp = ebp;
 	ctx.edi = edi;
@@ -372,26 +384,32 @@ throw_exception (unsigned long eax, unsigned long ecx, unsigned long edx, unsign
 	ctx.ecx = ecx;
 	ctx.eax = eax;
 
-	if (mono_debugger_throw_exception ((gpointer)(eip - 5), (gpointer)esp, exc)) {
-		/*
-		 * The debugger wants us to stop on the `throw' instruction.
-		 * By the time we get here, it already inserted a breakpoint on
-		 * eip - 5 (which is the address of the call).
-		 */
-		ctx.eip = eip - 5;
-		ctx.esp = esp + sizeof (gpointer);
-		restore_context (&ctx);
-		g_assert_not_reached ();
-	}
-
-	/* adjust eip so that it point into the call instruction */
-	ctx.eip -= 1;
-
 	if (mono_object_isinst (exc, mono_defaults.exception_class)) {
 		MonoException *mono_ex = (MonoException*)exc;
 		if (!rethrow)
 			mono_ex->stack_trace = NULL;
 	}
+
+	if (mono_debug_using_mono_debugger ()) {
+		guint8 buf [16], *code;
+
+		mono_breakpoint_clean_code (NULL, (gpointer)eip, 8, buf, sizeof (buf));
+		code = buf + 8;
+
+		if (buf [3] == 0xe8) {
+			MonoContext ctx_cp = ctx;
+			ctx_cp.eip = eip - 5;
+
+			if (mono_debugger_handle_exception (&ctx_cp, exc)) {
+				restore_context (&ctx_cp);
+				g_assert_not_reached ();
+			}
+		}
+	}
+
+	/* adjust eip so that it point into the call instruction */
+	ctx.eip -= 1;
+
 	mono_handle_exception (&ctx, exc, (gpointer)eip, FALSE);
 	restore_context (&ctx);
 
@@ -405,9 +423,14 @@ get_throw_exception (gboolean rethrow)
 
 	start = code = mono_global_codeman_reserve (64);
 
+	/* 
+	 * Align the stack on apple, since we push 10 args + the return address, and the
+	 * caller pushed 8 bytes.
+	 */
+	x86_alu_reg_imm (code, X86_SUB, X86_ESP, 4);
 	x86_push_reg (code, X86_ESP);
-	x86_push_membase (code, X86_ESP, 4); /* IP */
-	x86_push_membase (code, X86_ESP, 12); /* exception */
+	x86_push_membase (code, X86_ESP, 8); /* IP */
+	x86_push_membase (code, X86_ESP, 16); /* exception */
 	x86_push_reg (code, X86_EBP);
 	x86_push_reg (code, X86_EDI);
 	x86_push_reg (code, X86_ESI);
@@ -531,7 +554,9 @@ mono_arch_get_throw_corlib_exception (void)
 	inited = 1;
 	code = start = mono_global_codeman_reserve (64);
 
-	x86_push_membase (code, X86_ESP, 4); /* token */
+	x86_mov_reg_membase (code, X86_EAX, X86_ESP, 4, 4); /* token */
+	x86_alu_reg_imm (code, X86_ADD, X86_EAX, MONO_TOKEN_TYPE_DEF);
+	x86_push_reg (code, X86_EAX);
 	x86_push_imm (code, mono_defaults.exception_class->image);
 	x86_call_code (code, mono_exception_from_token);
 	x86_alu_reg_imm (code, X86_ADD, X86_ESP, 8);
@@ -563,8 +588,7 @@ mono_arch_get_throw_corlib_exception (void)
  */
 MonoJitInfo *
 mono_arch_find_jit_info (MonoDomain *domain, MonoJitTlsData *jit_tls, MonoJitInfo *res, MonoJitInfo *prev_ji, MonoContext *ctx, 
-			 MonoContext *new_ctx, char **trace, MonoLMF **lmf, int *native_offset,
-			 gboolean *managed)
+			 MonoContext *new_ctx, MonoLMF **lmf, gboolean *managed)
 {
 	MonoJitInfo *ji;
 	gpointer ip = MONO_CONTEXT_GET_IP (ctx);
@@ -573,7 +597,7 @@ mono_arch_find_jit_info (MonoDomain *domain, MonoJitTlsData *jit_tls, MonoJitInf
 	if (prev_ji && (ip > prev_ji->code_start && ((guint8*)ip < ((guint8*)prev_ji->code_start) + prev_ji->code_size)))
 		ji = prev_ji;
 	else
-		ji = mono_jit_info_table_find (domain, ip);
+		ji = mini_jit_info_table_find (domain, ip);
 
 	if (managed)
 		*managed = FALSE;
@@ -621,7 +645,7 @@ mono_arch_find_jit_info (MonoDomain *domain, MonoJitTlsData *jit_tls, MonoJitInf
 
 		if (*lmf && (MONO_CONTEXT_GET_BP (ctx) >= (gpointer)(*lmf)->ebp)) {
 			/* remove any unused lmf */
-			*lmf = (*lmf)->previous_lmf;
+			*lmf = (gpointer)(((guint32)(*lmf)->previous_lmf) & ~1);
 		}
 
 		/* Pop EBP and the return address */
@@ -643,9 +667,9 @@ mono_arch_find_jit_info (MonoDomain *domain, MonoJitTlsData *jit_tls, MonoJitInf
 		
 		*new_ctx = *ctx;
 
-		if ((ji = mono_jit_info_table_find (domain, (gpointer)(*lmf)->eip))) {
+		if ((ji = mini_jit_info_table_find (domain, (gpointer)(*lmf)->eip))) {
 		} else {
-			if (!(*lmf)->method)
+			if (!((guint32)((*lmf)->previous_lmf) & 1))
 				/* Top LMF entry */
 				return (gpointer)-1;
 			/* Trampoline lmf frame */
@@ -658,11 +682,29 @@ mono_arch_find_jit_info (MonoDomain *domain, MonoJitTlsData *jit_tls, MonoJitInf
 		new_ctx->ebx = (*lmf)->ebx;
 		new_ctx->ebp = (*lmf)->ebp;
 		new_ctx->eip = (*lmf)->eip;
-		/* the lmf is always stored on the stack, so the following
-		 * expression points to a stack location which can be used as ESP */
-		new_ctx->esp = (unsigned long)&((*lmf)->eip);
 
-		*lmf = (*lmf)->previous_lmf;
+		/* Check if we are in a trampoline LMF frame */
+		if ((guint32)((*lmf)->previous_lmf) & 1) {
+			/* lmf->esp is set by the trampoline code */
+			new_ctx->esp = (*lmf)->esp;
+
+			/* Pop arguments off the stack */
+			/* FIXME: Handle the delegate case too ((*lmf)->method == NULL) */
+			/* FIXME: Handle the IMT/vtable case too */
+			if ((*lmf)->method && (*lmf)->method != MONO_FAKE_IMT_METHOD && (*lmf)->method != MONO_FAKE_VTABLE_METHOD) {
+				MonoMethod *method = (*lmf)->method;
+				MonoJitArgumentInfo *arg_info = g_newa (MonoJitArgumentInfo, mono_method_signature (method)->param_count + 1);
+
+				guint32 stack_to_pop = mono_arch_get_argument_info (mono_method_signature (method), mono_method_signature (method)->param_count, arg_info);
+				new_ctx->esp += stack_to_pop;
+			}
+		}
+		else
+			/* the lmf is always stored on the stack, so the following
+			 * expression points to a stack location which can be used as ESP */
+			new_ctx->esp = (unsigned long)&((*lmf)->eip);
+
+		*lmf = (gpointer)(((guint32)(*lmf)->previous_lmf) & ~1);
 
 		return ji ? ji : res;
 	}
@@ -761,9 +803,115 @@ mono_arch_handle_exception (void *sigctx, gpointer obj, gboolean test_only)
 
 	mono_arch_sigctx_to_monoctx (sigctx, &mctx);
 
+	if (mono_debugger_handle_exception (&mctx, (MonoObject *)obj))
+		return TRUE;
+
 	mono_handle_exception (&mctx, obj, (gpointer)mctx.eip, test_only);
 
 	mono_arch_monoctx_to_sigctx (&mctx, sigctx);
 
 	return TRUE;
 }
+
+static void
+restore_soft_guard_pages (void)
+{
+	MonoJitTlsData *jit_tls = TlsGetValue (mono_jit_tls_id);
+	if (jit_tls->stack_ovf_guard_base)
+		mono_mprotect (jit_tls->stack_ovf_guard_base, jit_tls->stack_ovf_guard_size, MONO_MMAP_NONE);
+}
+
+/* 
+ * this function modifies mctx so that when it is restored, it
+ * won't execcute starting at mctx.eip, but in a function that
+ * will restore the protection on the soft-guard pages and return back to
+ * continue at mctx.eip.
+ */
+static void
+prepare_for_guard_pages (MonoContext *mctx)
+{
+	gpointer *sp;
+	sp = (gpointer)(mctx->esp);
+	sp -= 1;
+	/* the resturn addr */
+	sp [0] = (gpointer)(mctx->eip);
+	mctx->eip = (unsigned long)restore_soft_guard_pages;
+	mctx->esp = (unsigned long)sp;
+}
+
+static void
+altstack_handle_and_restore (void *sigctx, gpointer obj, gboolean stack_ovf)
+{
+	void (*restore_context) (MonoContext *);
+	MonoContext mctx;
+
+	restore_context = mono_arch_get_restore_context ();
+	mono_arch_sigctx_to_monoctx (sigctx, &mctx);
+
+	if (mono_debugger_handle_exception (&mctx, (MonoObject *)obj)) {
+		if (stack_ovf)
+			prepare_for_guard_pages (&mctx);
+		restore_context (&mctx);
+	}
+
+	mono_handle_exception (&mctx, obj, (gpointer)mctx.eip, FALSE);
+	if (stack_ovf)
+		prepare_for_guard_pages (&mctx);
+	restore_context (&mctx);
+}
+
+void
+mono_arch_handle_altstack_exception (void *sigctx, gpointer fault_addr, gboolean stack_ovf)
+{
+#ifdef MONO_ARCH_USE_SIGACTION
+	MonoException *exc = NULL;
+	ucontext_t *ctx = (ucontext_t*)sigctx;
+	MonoJitInfo *ji = mini_jit_info_table_find (mono_domain_get (), (gpointer)UCONTEXT_REG_EIP (ctx));
+	gpointer *sp;
+	int frame_size;
+
+	/* if we didn't find a managed method for the ip address and it matches the fault
+	 * address, we assume we followed a broken pointer during an indirect call, so
+	 * we try the lookup again with the return address pushed on the stack
+	 */
+	if (!ji && fault_addr == (gpointer)UCONTEXT_REG_EIP (ctx)) {
+		glong *sp = (gpointer)UCONTEXT_REG_ESP (ctx);
+		ji = mini_jit_info_table_find (mono_domain_get (), (gpointer)sp [0]);
+		if (ji)
+			UCONTEXT_REG_EIP (ctx) = sp [0];
+	}
+	if (stack_ovf)
+		exc = mono_domain_get ()->stack_overflow_ex;
+	if (!ji)
+		mono_handle_native_sigsegv (SIGSEGV, sigctx);
+	/* setup a call frame on the real stack so that control is returned there
+	 * and exception handling can continue.
+	 * If this was a stack overflow the caller already ensured the stack pages
+	 * needed have been unprotected.
+	 * The frame looks like:
+	 *   ucontext struct
+	 *   test_only arg
+	 *   exception arg
+	 *   ctx arg
+	 *   return ip
+	 */
+	frame_size = sizeof (ucontext_t) + sizeof (gpointer) * 4;
+	frame_size += 15;
+	frame_size &= ~15;
+	sp = (gpointer)(UCONTEXT_REG_ESP (ctx) & ~15);
+	sp = (gpointer)((char*)sp - frame_size);
+	/* the incoming arguments are aligned to 16 bytes boundaries, so the return address IP
+	 * goes at sp [-1]
+	 */
+	sp [-1] = (gpointer)UCONTEXT_REG_EIP (ctx);
+	sp [0] = sp + 4;
+	sp [1] = exc;
+	sp [2] = (gpointer)stack_ovf;
+	/* may need to adjust pointers in the new struct copy, depending on the OS */
+	memcpy (sp + 4, ctx, sizeof (ucontext_t));
+	/* at the return form the signal handler execution starts in altstack_handle_and_restore() */
+	UCONTEXT_REG_EIP (ctx) = (unsigned long)altstack_handle_and_restore;
+	UCONTEXT_REG_ESP (ctx) = (unsigned long)(sp - 1);
+#endif
+}
+
